

/*
 *  Author: Arvin Schnell
 */


#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <iostream>

using std::cerr;

#include "pcm-alsa.h"


ALSA::ALSA (const char* name, pcm_mode_t mode, pcm_format_t format,
	    int channels, int rate, int buffer_size)
    : PCM (name, mode, format, channels, rate, buffer_size),
      handle (0)
{
    if (channels <= 0) {
	cerr << "error: illegal number of channels\n";
	return;
    }

    int ret;

    snd_pcm_stream_t stream = mode == PLAY ? SND_PCM_STREAM_PLAYBACK :
	SND_PCM_STREAM_CAPTURE;
    ret = snd_pcm_open (&handle, name, stream, SND_PCM_NONBLOCK);
    if (ret < 0) {
	cerr << "error: can't open PCM device " << name << '\n';
	handle = 0;
	return;
    }

    ret = snd_pcm_nonblock (handle, 0);
    if (ret < 0) {
	cerr << "error: can't set blocking mode.\n";
	return;
    }

    snd_pcm_hw_params_t* hwparams;
    snd_pcm_hw_params_alloca (&hwparams);

    ret = snd_pcm_hw_params_any (handle, hwparams);
    if (ret < 0) {
	cerr << "error: can't configure the PCM device\n";
	return;
    }

    ret = snd_pcm_hw_params_set_access (handle, hwparams,
					SND_PCM_ACCESS_RW_INTERLEAVED);
    if (ret < 0) {
	cerr << "error: can't set access\n";
	return;
    }

    snd_pcm_format_t format_array[3] = { SND_PCM_FORMAT_S8, SND_PCM_FORMAT_S16,
					 SND_PCM_FORMAT_S32 };
    ret = snd_pcm_hw_params_set_format (handle, hwparams, format_array[format]);
    if (ret < 0) {
	cerr << "error: can't set format (resolution)\n";
	return;
    }

    ret = snd_pcm_hw_params_set_rate_near (handle, hwparams, rate, 0);
    if (ret < 0) {
	cerr << "error: can't set rate\n";
	return;
    }

    if (abs (rate - ret) > 1) {  // we can tolerate 1 Hz difference
	cerr << "error: can't set requested rate\n";
	cerr << "requested rate: " << rate << " Hz, actual rate: "
	     << ret << " Hz\n";
	return;
    }

    ret = snd_pcm_hw_params_set_channels (handle, hwparams, channels);
    if (ret < 0) {
	cerr << "error: can't set number of channels\n";
	return;
    }

    ret = snd_pcm_hw_params_set_periods (handle, hwparams, 2, 0);
    if (ret < 0) {
	cerr << "error: can't set number of periods\n";
	return;
    }

    ret = snd_pcm_hw_params_set_buffer_size (handle, hwparams, buffer_size);
    if (ret < 0) {
	cerr << "error: can't set buffer size\n";
	return;
    }

    ret = snd_pcm_hw_params (handle, hwparams);
    if (ret < 0) {
	cerr << "error: can't set HW params\n";
	return;
    }

    ok = true;
}


ALSA::~ALSA ()
{
    if (handle)
	snd_pcm_close (handle);
}


void
ALSA::info ()
{
    cerr << "driver: alsa 0.9\n";
    PCM::info ();
}


bool
ALSA::read (void* buffer, int frames)
{
    int ret = snd_pcm_readi (handle, buffer, frames);

    if (ret < 0) {
	if (ret == -EPIPE) {
	    cerr << "warning: buffer overrun\n";
	    snd_pcm_prepare (handle);
	} else {
	    cerr << "error: " << strerror (-ret) << '\n';
	    return false;
	}
    }

    return true;
}


bool
ALSA::write (void* buffer, int frames)
{
    int ret = snd_pcm_writei (handle, buffer, frames);

    if (ret < 0) {
	if (ret == -EPIPE) {
	    cerr << "warning: buffer underrun\n";
	    snd_pcm_prepare (handle);
	} else {
	    cerr << "error: " << strerror (-ret) << '\n';
	    return false;
	}
    }

    return true;
}
